import asyncio
import importlib
import inspect
import logging
import os
import sys
import time
from collections.abc import Generator
from contextlib import asynccontextmanager
from enum import Enum
from multiprocessing import Lock
from pathlib import Path
from threading import Thread
from typing import Any, AsyncGenerator, Dict, Optional, Set, Union

from anyio import Semaphore, to_thread
from common.patches import apply_patches
from common.retry import retry
from fastapi import HTTPException
from shared.secrets_resolver import SecretsResolver

MODEL_BASENAME = "model"

NUM_LOAD_RETRIES = int(os.environ.get("NUM_LOAD_RETRIES_TRUSS", "1"))
STREAMING_RESPONSE_QUEUE_READ_TIMEOUT_SECS = 60
DEFAULT_PREDICT_CONCURRENCY = 1


class DeferredSemaphoreManager:
    """
    Helper class for supported deferred semaphore release.
    """

    def __init__(self, semaphore: Semaphore):
        self.semaphore = semaphore
        self.deferred = False

    def defer(self):
        """
        Track that this semaphore is to be deferred, and return
        a release method that the context block can use to release
        the semaphore.
        """
        self.deferred = True

        return self.semaphore.release


@asynccontextmanager
async def deferred_semaphore(semaphore: Semaphore):
    """
    Context manager that allows deferring the release of a semaphore.
    It yields a DeferredSemaphoreManager -- in your use of this context manager,
    if you call DeferredSemaphoreManager.defer(), you will get back a function that releases
    the semaphore that you must call.
    """
    semaphore_manager = DeferredSemaphoreManager(semaphore)
    await semaphore.acquire()

    try:
        yield semaphore_manager
    finally:
        if not semaphore_manager.deferred:
            semaphore.release()


class ModelWrapper:
    class Status(Enum):
        NOT_READY = 0
        LOADING = 1
        READY = 2
        FAILED = 3

    def __init__(self, config: Dict):
        self._config = config
        self._logger = logging.getLogger()
        self.name = MODEL_BASENAME
        self.ready = False
        self._load_lock = Lock()
        self._status = ModelWrapper.Status.NOT_READY
        self._predict_semaphore = Semaphore(
            self._config.get("runtime", {}).get(
                "predict_concurrency", DEFAULT_PREDICT_CONCURRENCY
            )
        )
        self._background_tasks: Set[asyncio.Task] = set()

    def load(self) -> bool:
        if self.ready:
            return self.ready

        # if we are already loading, block on aquiring the lock;
        # this worker will return 503 while the worker with the lock is loading
        with self._load_lock:
            self._status = ModelWrapper.Status.LOADING

            self._logger.info("Executing model.load()...")

            try:
                start_time = time.perf_counter()
                self.try_load()
                self.ready = True
                self._status = ModelWrapper.Status.READY
                self._logger.info(
                    f"Completed model.load() execution in {_elapsed_ms(start_time)} ms"
                )

                return self.ready
            except Exception:
                self._logger.exception("Exception while loading model")
                self._status = ModelWrapper.Status.FAILED

        return self.ready

    def start_load(self):
        if self.should_load():
            thread = Thread(target=self.load)
            thread.start()

    def load_failed(self) -> bool:
        return self._status == ModelWrapper.Status.FAILED

    def should_load(self) -> bool:
        # don't retry failed loads
        return not self._status == ModelWrapper.Status.FAILED and not self.ready

    def try_load(self):
        data_dir = Path("data")
        data_dir.mkdir(exist_ok=True)

        if "bundled_packages_dir" in self._config:
            bundled_packages_path = Path("/packages")
            if bundled_packages_path.exists():
                sys.path.append(str(bundled_packages_path))
        model_module_name = str(
            Path(self._config["model_class_filename"]).with_suffix("")
        )
        module = importlib.import_module(
            f"{self._config['model_module_dir']}.{model_module_name}"
        )
        model_class = getattr(module, self._config["model_class_name"])
        model_class_signature = inspect.signature(model_class)
        model_init_params = {}
        if _signature_accepts_keyword_arg(model_class_signature, "config"):
            model_init_params["config"] = self._config
        if _signature_accepts_keyword_arg(model_class_signature, "data_dir"):
            model_init_params["data_dir"] = data_dir
        if _signature_accepts_keyword_arg(model_class_signature, "secrets"):
            model_init_params["secrets"] = SecretsResolver.get_secrets(self._config)
        apply_patches(
            self._config.get("apply_library_patches", True),
            self._config["requirements"],
        )
        self._model = model_class(**model_init_params)

        if hasattr(self._model, "load"):
            retry(
                self._model.load,
                NUM_LOAD_RETRIES,
                self._logger.warn,
                "Failed to load model.",
                gap_seconds=1.0,
            )

    async def preprocess(
        self,
        payload: Any,
        headers: Optional[Dict[str, str]] = None,
    ) -> Any:
        if not hasattr(self._model, "preprocess"):
            return payload

        if inspect.iscoroutinefunction(self._model.preprocess):
            return await _intercept_exceptions_async(self._model.preprocess)(payload)
        else:
            return await to_thread.run_sync(
                _intercept_exceptions_sync(self._model.preprocess), payload
            )

    async def predict(
        self,
        payload: Any,
        headers: Optional[Dict[str, str]] = None,
    ) -> Any:
        # It's possible for the user's predict function to be a:
        #   1. Generator function (function that returns a generator)
        #   2. Async generator (function that returns async generator)
        # In these cases, just return the generator or async generator,
        # as we will be propagating these up. No need for await at this point.
        #   3. Coroutine -- in this case, await the predict function as it is async
        #   4. Normal function -- in this case, offload to a separate thread to prevent
        #      blocking the main event loop
        if inspect.isasyncgenfunction(
            self._model.predict
        ) or inspect.isgeneratorfunction(self._model.predict):
            return self._model.predict(payload)

        if inspect.iscoroutinefunction(self._model.predict):
            return await _intercept_exceptions_async(self._model.predict)(payload)

        return await to_thread.run_sync(
            _intercept_exceptions_sync(self._model.predict), payload
        )

    async def postprocess(
        self,
        response: Any,
        headers: Optional[Dict[str, str]] = None,
    ) -> Any:
        # Similar to the predict function, it is possible for postprocess
        # to return either a generator or async generator, in which case
        # just return the generator.
        #
        # It can also return a coroutine or just be a function, in which
        # case either await, or offload to a thread respectively.
        if not hasattr(self._model, "postprocess"):
            return response

        if inspect.isasyncgenfunction(
            self._model.postprocess
        ) or inspect.isgeneratorfunction(self._model.postprocess):
            return self._model.postprocess(response)

        if inspect.iscoroutinefunction(self._model.postprocess):
            return await _intercept_exceptions_async(self._model.postprocess)(response)

        return await to_thread.run_sync(
            _intercept_exceptions_sync(self._model.postprocess), response
        )

    async def write_response_to_queue(
        self, queue: asyncio.Queue, generator: AsyncGenerator
    ):
        try:
            async for chunk in generator:
                await queue.put(ResponseChunk(chunk))
        except Exception as e:
            self._logger.exception("Exception while reading stream response: " + str(e))
        finally:
            await queue.put(None)

    async def __call__(
        self, body: Any, headers: Optional[Dict[str, str]] = None
    ) -> Union[Dict, Generator]:
        """Method to call predictor or explainer with the given input.

        Args:
            body (Any): Request payload body.
            headers (Dict): Request headers.

        Returns:
            Dict: Response output from preprocess -> predictor -> postprocess
            Generator: In case of streaming response
        """

        payload = await self.preprocess(body, headers)

        async with deferred_semaphore(self._predict_semaphore) as semaphore_manager:
            response = await self.predict(payload, headers)

            # Streaming cases
            if inspect.isgenerator(response) or inspect.isasyncgen(response):
                if hasattr(self._model, "postprocess"):
                    logging.warning(
                        "Predict returned a streaming response, while a postprocess is defined."
                        "Note that in this case, the postprocess will run within the predict lock."
                    )

                    response = await self.postprocess(response)

                async_generator = _force_async_generator(response)

                if headers and headers.get("accept") == "application/json":
                    # In the case of a streaming response, consume stream
                    # if the http accept header is set, and json is requested.
                    return await _convert_streamed_response_to_string(async_generator)

                # To ensure that a partial read from a client does not cause the semaphore
                # to stay claimed, we immediately write all of the data from the stream to a
                # queue. We then return a new generator that reads from the queue, and then
                # exit the semaphore block.
                response_queue: asyncio.Queue = asyncio.Queue()

                # This task will be triggered and run in the background.
                task = asyncio.create_task(
                    self.write_response_to_queue(response_queue, async_generator)
                )

                # We add the task to the ModelWrapper instance to ensure it does
                # not get garbage collected after the predict method completes,
                # and continues running.
                self._background_tasks.add(task)

                # Defer the release of the semaphore until the write_response_to_queue
                # task.
                semaphore_release_function = semaphore_manager.defer()
                task.add_done_callback(lambda _: semaphore_release_function())
                task.add_done_callback(self._background_tasks.discard)

                async def _response_generator():
                    while True:
                        chunk = await asyncio.wait_for(
                            response_queue.get(),
                            timeout=STREAMING_RESPONSE_QUEUE_READ_TIMEOUT_SECS,
                        )
                        if chunk is None:
                            return
                        yield chunk.value

                return _response_generator()

        processed_response = await self.postprocess(response)
        return processed_response


class ResponseChunk:
    def __init__(self, value):
        self.value = value


async def _convert_streamed_response_to_string(response: AsyncGenerator):
    return "".join([str(chunk) async for chunk in response])


def _force_async_generator(gen: Union[Generator, AsyncGenerator]) -> AsyncGenerator:
    """
    Takes a generator, and converts it into an async generator if it is not already.
    """
    if inspect.isasyncgen(gen):
        return gen

    async def _convert_generator_to_async():
        """
        Runs each iteration of the generator in an offloaded thread, to ensure
        the main loop is not blocked, and yield to create an async generator.
        """
        FINAL_GENERATOR_VALUE = object()
        while True:
            # Note that this is the equivalent of running:
            # next(gen, FINAL_GENERATOR_VALUE) on a separate thread,
            # ensuring that if there is anything blocking in the generator,
            # it does not block the main loop.
            chunk = await to_thread.run_sync(next, gen, FINAL_GENERATOR_VALUE)
            if chunk == FINAL_GENERATOR_VALUE:
                break
            yield chunk

    return _convert_generator_to_async()


def _signature_accepts_keyword_arg(signature: inspect.Signature, kwarg: str) -> bool:
    return kwarg in signature.parameters or _signature_accepts_kwargs(signature)


def _signature_accepts_kwargs(signature: inspect.Signature) -> bool:
    for param in signature.parameters.values():
        if param.kind == inspect.Parameter.VAR_KEYWORD:
            return True
    return False


def _elapsed_ms(since_micro_seconds: float) -> int:
    return int((time.perf_counter() - since_micro_seconds) * 1000)


def _handle_exception():
    # Note that logger.exception logs the stacktrace, such that the user can
    # debug this error from the logs.
    logging.exception("Internal Server Error")
    raise HTTPException(status_code=500, detail="Internal Server Error")


def _intercept_exceptions_sync(func):
    def inner(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception:
            _handle_exception()

    return inner


def _intercept_exceptions_async(func):
    async def inner(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except Exception:
            _handle_exception()

    return inner



from typing import Any, List, Type, Union

import numpy as np
from pydantic import BaseModel
from utils.errors import UnsupportedTypeError

# This is a broad type that includes built-in Python types and types that might be
# returned by Pydantic functions like conlist.
FieldAnnotationType = Union[
    Type[int], Type[float], Type[str], Type[bool], Type[dict], List[Any]
]

# Dictionary to specify the numpy dtype when converting from Python types to Triton tensors
PYTHON_TYPE_TO_NP_DTYPE = {
    int: np.int32,
    float: np.float32,
    str: np.dtype(object),
    bool: np.bool_,
    dict: np.dtype(object),
}


def is_conlist(pydantic_field_annotation: FieldAnnotationType) -> bool:
    """Checks if a Pydantic field annotation is a conlist."""
    # Annotations that are not directly a Python type have an __origin__ attribute
    if not hasattr(pydantic_field_annotation, "__origin__"):
        return False
    return pydantic_field_annotation.__origin__ == list


def get_pydantic_field_type(pydantic_field_annotation: FieldAnnotationType) -> str:
    """
    Returns the type of a Pydantic field via it's annotation. The annotation can be a
    type or a type with constraints (e.g. conlist). For the following types on a Pydantic
    field, the pydantic_field_annotation is

    - str: str
    - int: int
    - float: float
    - bool: bool
    - list: list
    - conlist: typing.List[...] where ... is the type of the list elements

    Args:
        pydantic_field (FieldAnnotationType): The Pydantic field.

    Returns:
        The type (str) of the Pydantic field.
    """
    if isinstance(pydantic_field_annotation, type):
        return pydantic_field_annotation.__name__
    elif is_conlist(pydantic_field_annotation):
        return "list"
    else:
        raise UnsupportedTypeError(
            f"Unsupported Pydantic field annotation type: {pydantic_field_annotation}"
        )


def inspect_pydantic_model(pydantic_class: Type[BaseModel]) -> list:
    """
    Inspects a Pydantic model and returns a list of fields and their types.

    Args:
        pydantic_class (Type[BaseModel]): The Pydantic model class.

    Returns:
        A list of dicts containing the field name and type.
    """
    return [
        {"param": field_name, "type": get_pydantic_field_type(field_info.annotation)}
        for field_name, field_info in pydantic_class.__fields__.items()  # type: ignore
    ]



from typing import Dict, Optional

from truss.remote.baseten.auth import AuthService
from truss.remote.truss_remote import TrussService
from truss.truss_handle import TrussHandle

DEFAULT_STREAM_ENCODING = "utf-8"


class BasetenService(TrussService):
    def __init__(
        self,
        model_id: str,
        model_version_id: str,
        is_draft: bool,
        api_key: str,
        service_url: str,
        truss_handle: Optional[TrussHandle] = None,
    ):
        super().__init__(is_draft=is_draft, service_url=service_url)
        self._model_id = model_id
        self._model_version_id = model_version_id
        self._auth_service = AuthService(api_key=api_key)
        self._truss_handle = truss_handle

    def is_live(self) -> bool:
        raise NotImplementedError

    def is_ready(self) -> bool:
        raise NotImplementedError

    @property
    def model_id(self) -> str:
        return self._model_id

    @property
    def model_version_id(self) -> str:
        return self._model_version_id

    @property
    def invocation_url(self) -> str:
        return f"{self._service_url}/predict"

    def predict(
        self,
        model_request_body: Dict,
    ):
        response = self._send_request(
            self.invocation_url, "POST", data=model_request_body, stream=True
        )

        if response.headers.get("transfer-encoding") == "chunked":
            # Case of streaming response, the backend does not set an encoding, so
            # manually decode to the contents to utf-8 here.
            def decode_content():
                for chunk in response.iter_content(
                    chunk_size=8192, decode_unicode=True
                ):
                    # Depending on the content-type of the response,
                    # iter_content will either emit a byte stream, or a stream
                    # of strings. Only decode in the bytes case.
                    if isinstance(chunk, bytes):
                        yield chunk.decode(response.encoding or DEFAULT_STREAM_ENCODING)
                    else:
                        yield chunk

            return decode_content()

        parsed_response = response.json()

        if "error" in parsed_response:
            # In the case that the model is in a non-ready state, the response
            # will be a json with an `error` key.
            return parsed_response

        return response.json()["model_output"]

    def authenticate(self) -> dict:
        return self._auth_service.authenticate().header()

    def logs_url(self, base_url: str) -> str:
        return f"{base_url}/models/{self._model_id}/logs/{self._model_version_id}"



import importlib
import inspect
import os
import sys
from pathlib import Path

import yaml
from shared.secrets_resolver import SecretsResolver

CONFIG_FILE = "config.yaml"

# This is where user training module will be mounted
TRAINING_CODE_PATH = "/train"
OUTPUT_PATH = "/output"
VARIABLES_PATH = "/variables"
VARIABLES_FILE = "variables.yaml"


def _signature_accepts_keyword_arg(signature: inspect.Signature, kwarg: str) -> bool:
    return kwarg in signature.parameters or _signature_accepts_kwargs(signature)


# todo: avoid duplication with model_wrapper
def _signature_accepts_kwargs(signature: inspect.Signature) -> bool:
    for param in signature.parameters.values():
        if param.kind == inspect.Parameter.VAR_KEYWORD:
            return True
    return False


def _add_bundled_packages_to_path(config) -> None:
    if "bundled_packages_dir" in config:
        bundled_packages_path = Path("/packages")
        if bundled_packages_path.exists():
            sys.path.append(str(bundled_packages_path))


def _load_train_class(config):
    if not Path(TRAINING_CODE_PATH).exists():
        raise ValueError("Training code not mounted")

    # This should be created via mounting, but if not then create it.
    if not Path(OUTPUT_PATH).exists():
        Path(OUTPUT_PATH).mkdir()

    sys.path.append(TRAINING_CODE_PATH)
    train_config = config["train"]
    training_module_name = str(
        Path(train_config["training_class_filename"]).with_suffix("")
    )
    train_module = importlib.import_module(f"{training_module_name}")
    return getattr(train_module, train_config["training_class_name"])


def _create_trainer(config):
    train_class = _load_train_class(config)
    train_class_signature = inspect.signature(train_class)
    train_init_params = {}
    if _signature_accepts_keyword_arg(train_class_signature, "config"):
        train_init_params["config"] = config
    if _signature_accepts_keyword_arg(train_class_signature, "output_dir"):
        train_init_params["output_dir"] = Path(OUTPUT_PATH)

    # Wire up secrets
    if _signature_accepts_keyword_arg(train_class_signature, "secrets"):
        train_init_params["secrets"] = SecretsResolver.get_secrets(config)

    # Wire up variables
    if _signature_accepts_keyword_arg(train_class_signature, "variables"):
        default_variables = {}
        if "train" in config and "variables" in config["train"]:
            default_variables = config["train"]["variables"]

        runtime_variables = {}
        vars_path = Path(VARIABLES_PATH) / VARIABLES_FILE
        if vars_path.exists():
            with vars_path.open() as vars_file:
                runtime_variables = yaml.safe_load(vars_file)

        variables = {
            **default_variables,
            **runtime_variables,
        }
        train_init_params["variables"] = variables
    return train_class(**train_init_params)


if __name__ == "__main__":
    with open(CONFIG_FILE, encoding="utf-8") as config_file:
        truss_config = yaml.safe_load(config_file)
        _add_bundled_packages_to_path(truss_config)
        sys.path.append(os.environ["APP_HOME"])
        trainer = _create_trainer(truss_config)
        if hasattr(trainer, "pre_train"):
            trainer.pre_train()

        # train is a required method, so no check
        trainer.train()

        if hasattr(trainer, "post_train"):
            trainer.post_train()



# This file doesn't test anything, but provides utilities for testing.
from unittest import mock


def model_supports_predict_proba():
    mock_not_predict_proba = mock.Mock(name="mock_not_predict_proba")
    mock_not_predict_proba.predict_proba.return_value = False

    mock_check_proba = mock.Mock(name="mock_check_proba")
    mock_check_proba.predict_proba.return_value = True
    mock_check_proba._check_proba.return_value = True

    mock_not_check_proba = mock.Mock(name="mock_not_check_proba")
    mock_not_check_proba.predict_proba.return_value = True
    mock_not_check_proba._check_proba.side_effect = AttributeError

    assert not model_supports_predict_proba(mock_not_predict_proba)
    assert model_supports_predict_proba(mock_check_proba)
    assert not model_supports_predict_proba(mock_not_check_proba)



import inspect
import json
import logging
import os
import sys
from functools import wraps
from pathlib import Path
from typing import Callable, Optional

import rich
import rich_click as click
import truss
from truss.cli.console import console
from truss.cli.create import ask_name, select_server_backend
from truss.remote.baseten.core import (
    ModelId,
    ModelIdentifier,
    ModelName,
    ModelVersionId,
)
from truss.remote.baseten.service import BasetenService
from truss.remote.remote_cli import inquire_model_name, inquire_remote_name
from truss.remote.remote_factory import USER_TRUSSRC_PATH, RemoteFactory
from truss.truss_config import ModelServer

logging.basicConfig(level=logging.INFO)


click.rich_click.COMMAND_GROUPS = {
    "truss": [
        {
            "name": "Main usage",
            "commands": ["init", "push", "watch", "predict"],
            "table_styles": {  # type: ignore
                "row_styles": ["green"],
            },
        },
        {
            "name": "Advanced Usage",
            "commands": ["image", "container", "cleanup"],
            "table_styles": {  # type: ignore
                "row_styles": ["yellow"],
            },
        },
    ]
}


def echo_output(f: Callable[..., object]):
    @wraps(f)
    def wrapper(*args, **kwargs):
        click.echo(f(*args, **kwargs))

    return wrapper


def error_handling(f: Callable[..., object]):
    @wraps(f)
    def wrapper(*args, **kwargs):
        try:
            f(*args, **kwargs)
        except click.UsageError as e:
            raise e  # You can re-raise the exception or handle it different
        except Exception as e:
            click.secho(f"ERROR: {e}", fg="red")

    return wrapper


def print_help() -> None:
    ctx = click.get_current_context()
    click.echo(ctx.get_help())


@click.group(name="truss", invoke_without_command=True)  # type: ignore
@click.pass_context
@click.version_option(truss.version())
def truss_cli(ctx) -> None:
    """truss: The simplest way to serve models in production"""
    if not ctx.invoked_subcommand:
        click.echo(ctx.get_help())


@click.group()
def container():
    """Subcommands for truss container"""
    pass


@click.group()
def image():
    """Subcommands for truss image"""
    pass


@truss_cli.command()
@click.argument("target_directory", required=True)
@click.option(
    "-b",
    "--backend",
    show_default=True,
    default=ModelServer.TrussServer.value,
    type=click.Choice([server.value for server in ModelServer]),
)
@error_handling
def init(target_directory, backend) -> None:
    """Create a new truss.

    TARGET_DIRECTORY: A Truss is created in this directory
    """
    if os.path.isdir(target_directory):
        raise click.ClickException(
            f'Error: Directory "{target_directory}" already exists and cannot be overwritten.'
        )
    tr_path = Path(target_directory)
    build_config = select_server_backend(ModelServer[backend])
    model_name = ask_name()
    truss.init(
        target_directory=target_directory,
        build_config=build_config,
        model_name=model_name,
    )
    click.echo(f"Truss {model_name} was created in {tr_path.absolute()}")


@image.command()  # type: ignore
@click.argument("build_dir")
@click.argument("target_directory", required=False)
@error_handling
def build_context(build_dir, target_directory: str) -> None:
    """
    Create a docker build context for a Truss.

    BUILD_DIR: Folder where image context is built for Truss

    TARGET_DIRECTORY: A Truss directory. If none, use current directory.
    """
    tr = _get_truss_from_directory(target_directory=target_directory)
    tr.docker_build_setup(build_dir=Path(build_dir))


@image.command()  # type: ignore
@click.argument("target_directory", required=False)
@click.argument("build_dir", required=False)
@error_handling
@click.option("--tag", help="Docker image tag")
def build(target_directory: str, build_dir: Path, tag) -> None:
    """
    Builds the docker image for a Truss.

    TARGET_DIRECTORY: A Truss directory. If none, use current directory.

    BUILD_DIR: Image context. If none, a temp directory is created.
    """
    tr = _get_truss_from_directory(target_directory=target_directory)
    if build_dir:
        build_dir = Path(build_dir)
    tr.build_serving_docker_image(build_dir=build_dir, tag=tag)


@image.command()  # type: ignore
@click.argument("target_directory", required=False)
@click.argument("build_dir", required=False)
@click.option("--tag", help="Docker build image tag")
@click.option("--port", type=int, default=8080, help="Local port used to run image")
@click.option(
    "--attach", is_flag=True, default=False, help="Flag for attaching the process"
)
@error_handling
def run(target_directory: str, build_dir: Path, tag, port, attach) -> None:
    """
    Runs the docker image for a Truss.

    TARGET_DIRECTORY: A Truss directory. If none, use current directory.

    BUILD_DIR: Image context. If none, a temp directory is created.
    """
    tr = _get_truss_from_directory(target_directory=target_directory)
    if build_dir:
        build_dir = Path(build_dir)
    urls = tr.get_urls_from_truss()
    if urls:
        click.confirm(
            f"Container already exists at {urls}. Are you sure you want to continue?"
        )
    tr.docker_run(build_dir=build_dir, tag=tag, local_port=port, detach=not attach)


@truss_cli.command()
@click.argument("target_directory", required=False, default=os.getcwd())
@click.option(
    "--remote",
    type=str,
    required=False,
    help="Name of the remote in .trussrc to patch changes to",
)
@click.option(
    "--logs",
    is_flag=True,
    show_default=True,
    default=False,
    help="Automatically open remote logs tab",
)
@error_handling
def watch(
    target_directory: str,
    remote: str,
    logs: bool,
) -> None:
    """
    Seamless remote development with truss

    TARGET_DIRECTORY: A Truss directory. If none, use current directory.
    """

    # TODO: ensure that provider support draft
    if not remote:
        remote = inquire_remote_name(RemoteFactory.get_available_config_names())

    remote_provider = RemoteFactory.create(remote=remote)

    tr = _get_truss_from_directory(target_directory=target_directory)
    model_name = tr.spec.config.model_name
    if not model_name:
        rich.print(
            "🧐 NoneType model_name provided in config.yaml. "
            "Please check that you have the correct model name in your config file."
        )
        sys.exit(1)

    service = remote_provider.get_service(model_identifier=ModelName(model_name))
    logs_url = remote_provider.get_remote_logs_url(service)
    rich.print(f"🪵  View logs for your deployment at {logs_url}")
    remote_provider.sync_truss_to_dev_version_by_name(model_name, target_directory)


def _extract_and_validate_model_identifier(
    target_directory: str,
    model_id: Optional[str],
    model_version_id: Optional[str],
    published: Optional[bool],
) -> ModelIdentifier:
    if published and (model_id or model_version_id):
        raise click.UsageError(
            "Cannot use --published with --model or --model-deployment."
        )

    model_identifier: ModelIdentifier
    if model_version_id:
        model_identifier = ModelVersionId(model_version_id)
    elif model_id:
        model_identifier = ModelId(model_id)
    else:
        tr = _get_truss_from_directory(target_directory=target_directory)
        model_name = tr.spec.config.model_name
        if not model_name:
            raise click.UsageError("Truss config is missing a model name.")
        model_identifier = ModelName(model_name)
    return model_identifier


def _extract_request_data(data: Optional[str], file: Optional[Path]):
    try:
        if data is not None:
            return json.loads(data)
        if file is not None:
            return json.loads(Path(file).read_text())
    except json.JSONDecodeError:
        raise click.UsageError("Request data must be valid json.")

    raise click.UsageError(
        "You must provide exactly one of '--data (-d)' or '--file (-f)' options."
    )


@truss_cli.command()
@click.option("--target-directory", required=False, help="Directory of truss")
@click.option(
    "--remote",
    type=str,
    required=False,
    help="Name of the remote in .trussrc to push to",
)
@click.option(
    "-d",
    "--data",
    type=str,
    required=False,
    help="String formatted as json that represents request",
)
@click.option(
    "-f",
    "--file",
    type=click.Path(exists=True),
    help="Path to json file containing the request",
)
@click.option(
    "--published",
    type=bool,
    is_flag=True,
    required=False,
    default=False,
    help="Call the published model deployment.",
)
@click.option(
    "--model-version",
    type=str,
    required=False,
    help="[DEPRECATED] Use --model-deployment instead, this will be removed in future release. ID of model deployment",
)
@click.option(
    "--model-deployment",
    type=str,
    required=False,
    help="ID of model deployment to call",
)
@click.option(
    "--model",
    type=str,
    required=False,
    help="ID of model to call",
)
@echo_output
def predict(
    target_directory: str,
    remote: str,
    data: Optional[str],
    file: Optional[Path],
    published: Optional[bool],
    model_version: Optional[str],
    model_deployment: Optional[str],
    model: Optional[str],
):
    """
    Calls the packaged model

    TARGET_DIRECTORY: A Truss directory. If none, use current directory.

    REQUEST: String formatted as json that represents request

    REQUEST_FILE: Path to json file containing the request
    """
    if not remote:
        remote = inquire_remote_name(RemoteFactory.get_available_config_names())

    remote_provider = RemoteFactory.create(remote=remote)

    if model_version:
        console.print(
            "[DEPRECATED] --model-version is deprecated, use --model-deployment instead.",
            style="yellow",
        )
        model_deployment = model_version

    model_identifier = _extract_and_validate_model_identifier(
        target_directory,
        model_id=model,
        model_version_id=model_deployment,
        published=published,
    )

    request_data = _extract_request_data(data=data, file=file)

    service = remote_provider.get_service(
        model_identifier=model_identifier, published=published
    )

    # Log deployment ID for Baseten models.
    if isinstance(service, BasetenService):
        rich.print(
            f"Calling predict on {'[cyan]development[/cyan] ' if service.is_draft else ''}"
            f"deployment ID {service.model_version_id}..."
        )

    result = service.predict(request_data)
    if inspect.isgenerator(result):
        for chunk in result:
            click.echo(chunk, nl=False)
        return
    rich.print_json(data=result)


@truss_cli.command()
@click.argument("target_directory", required=False, default=os.getcwd())
@click.option(
    "--remote",
    type=str,
    required=False,
    help="Name of the remote in .trussrc to push to",
)
@click.option("--model-name", type=str, required=False, help="Name of the model")
@click.option(
    "--publish",
    type=bool,
    is_flag=True,
    required=False,
    default=False,
    help=(
        "Push the truss as a published deployment. If no production "
        "deployment exists, promote the truss to production "
        "after deploy completes."
    ),
)
@click.option(
    "--promote",
    type=bool,
    is_flag=True,
    required=False,
    default=False,
    help=(
        "Push the truss as a published deployment. Even if a production "
        "deployment exists, promote the truss to production "
        "after deploy completes."
    ),
)
@click.option(
    "--preserve-previous-production-deployment",
    type=bool,
    is_flag=True,
    required=False,
    default=False,
    help=(
        "Preserve the previous production deployment's autoscaling setting. When not specified, "
        "the previous production deployment will be updated to allow it to scale to zero. "
        "Can only be use in combination with --promote option"
    ),
)
@click.option(
    "--trusted",
    type=bool,
    is_flag=True,
    required=False,
    default=False,
    help="Trust truss with hosted secrets.",
)
@click.option(
    "--deployment-name",
    type=str,
    required=False,
    help=(
        "Name of the deployment created by the push. Can only be "
        "used in combination with --publish or --promote."
    ),
)
@error_handling
def push(
    target_directory: str,
    remote: str,
    model_name: str,
    publish: bool = False,
    trusted: bool = False,
    promote: bool = False,
    preserve_previous_production_deployment: bool = False,
    deployment_name: Optional[str] = None,
) -> None:
    """
    Pushes a truss to a TrussRemote.

    TARGET_DIRECTORY: A Truss directory. If none, use current directory.

    """
    if not remote:
        remote = inquire_remote_name(RemoteFactory.get_available_config_names())

    remote_provider = RemoteFactory.create(remote=remote)

    tr = _get_truss_from_directory(target_directory=target_directory)

    model_name = model_name or tr.spec.config.model_name
    if not model_name:
        model_name = inquire_model_name()

    # Write model name to config if it's not already there
    if model_name != tr.spec.config.model_name:
        tr.spec.config.model_name = model_name
        tr.spec.config.write_to_yaml_file(tr.spec.config_path, verbose=False)

    # TODO(Abu): This needs to be refactored to be more generic
    service = remote_provider.push(
        tr,
        model_name,
        publish=publish,
        trusted=trusted,
        promote=promote,
        preserve_previous_prod_deployment=preserve_previous_production_deployment,
        deployment_name=deployment_name,
    )  # type: ignore

    click.echo(f"✨ Model {model_name} was successfully pushed ✨")

    if service.is_draft:
        draft_model_text = """
|---------------------------------------------------------------------------------------|
| Your model has been deployed as a development model. Development models allow you to  |
| iterate quickly during the deployment process.                                        |
|                                                                                       |
| When you are ready to publish your deployed model as a new deployment,                |
| pass `--publish` to the `truss push` command. To monitor changes to your model and    |
| rapidly iterate, run the `truss watch` command.                                       |
|                                                                                       |
|---------------------------------------------------------------------------------------|
"""

        click.echo(draft_model_text)

    # Log a warning if using secrets without --trusted.
    # TODO(helen): this could be moved to a separate function that includes more config checks.
    if tr.spec.config.secrets and not trusted:
        not_trusted_text = """Warning: your Truss has secrets but was not pushed with --trusted.
Please push with --trusted to grant access to secrets.
"""
        console.print(not_trusted_text, style="red")

    if promote:
        promotion_text = """Your Truss has been deployed as a production model. After it successfully deploys,
it will become the next production deployment of your model."""
        console.print(promotion_text, style="green")

    logs_url = remote_provider.get_remote_logs_url(service)  # type: ignore[attr-defined]
    rich.print(f"🪵  View logs for your deployment at {logs_url}")


@truss_cli.command()
def configure():
    # Read the original file content
    with open(USER_TRUSSRC_PATH, "r") as f:
        original_content = f.read()

    # Open the editor and get the modified content
    edited_content = click.edit(original_content)

    # If the content was modified, save it
    if edited_content is not None and edited_content != original_content:
        with open(USER_TRUSSRC_PATH, "w") as f:
            f.write(edited_content)
            click.echo(f"Changes saved to {USER_TRUSSRC_PATH}")
    else:
        click.echo("No changes made.")


@container.command()  # type: ignore
@click.argument("target_directory", required=False)
@error_handling
def logs(target_directory) -> None:
    """
    Get logs in a container is running for a truss

    TARGET_DIRECTORY: A Truss directory. If none, use current directory.
    """
    for log in _get_truss_from_directory(
        target_directory=target_directory
    ).serving_container_logs():
        click.echo(log)


@container.command()  # type: ignore
@click.argument("target_directory", required=False)
def kill(target_directory: str) -> None:
    """
    Kills containers related to truss.

    TARGET_DIRECTORY: A Truss directory. If none, use current directory.
    """
    tr = _get_truss_from_directory(target_directory=target_directory)
    tr.kill_container()


@container.command()  # type: ignore
def kill_all() -> None:
    "Kills all truss containers that are not manually persisted"
    truss.kill_all()


@truss_cli.command()
@error_handling
def cleanup() -> None:
    """
    Clean up truss data.

    Truss creates temporary directories for various operations
    such as for building docker images. This command clears
    that data to free up disk space.
    """
    truss.build.cleanup()


def _get_truss_from_directory(target_directory: Optional[str] = None):
    """Gets Truss from directory. If none, use the current directory"""
    if target_directory is None:
        target_directory = os.getcwd()
    return truss.load(target_directory)


truss_cli.add_command(container)
truss_cli.add_command(image)

if __name__ == "__main__":
    truss_cli()



import importlib
import sys
from contextlib import contextmanager
from importlib.machinery import PathFinder
from pathlib import Path
from typing import List, Optional


class TrussModuleFinder(PathFinder):
    _truss_dir: str
    _bundled_packages_dir_name: Optional[str]
    _external_packages_dirs: Optional[List[str]]
    _truss_module_name: Optional[str]

    @classmethod
    def set_model_truss_dirs(
        cls,
        truss_dir: str,
        truss_module_name: Optional[str] = None,
        bundled_packages_dir_name: Optional[str] = None,
        external_packages_dirs: Optional[List[str]] = None,
    ):
        cls._truss_dir = truss_dir
        cls._truss_module_name = truss_module_name
        cls._bundled_packages_dir_name = bundled_packages_dir_name
        cls._external_packages_dirs = external_packages_dirs
        cls.add_to_meta_path()

    @classmethod
    def find_spec(cls, fullname, path=None, target=None):
        top_level_module_name = fullname.split(".")[0]
        if top_level_module_name == cls._truss_module_name:
            truss_modules_path = [cls._truss_dir]
        else:
            truss_modules_path = [
                str(Path(cls._truss_dir) / cls._bundled_packages_dir_name),
            ]
            if cls._external_packages_dirs:
                truss_modules_path += [
                    str(Path(cls._truss_dir) / d) for d in cls._external_packages_dirs
                ]
        if not path:
            path = truss_modules_path
        else:
            path = list(path) + truss_modules_path
        return PathFinder.find_spec(fullname, path, target)

    @classmethod
    def add_to_meta_path(cls):
        if cls not in sys.meta_path:
            sys.meta_path.append(cls)

    @classmethod
    def remove_from_meta_path(cls):
        if cls in sys.meta_path:
            del sys.meta_path[sys.meta_path.index(cls)]


class TrussModuleLoader:
    @staticmethod
    def import_truss_module(name):
        """
        Custom import implementation for loading model or training module.

        Based on https://docs.python.org/3/library/importlib.html#approximating-importlib-import-module
        """
        absolute_name = importlib.util.resolve_name(name, None)
        path = None
        if "." in absolute_name:
            parent_name, _, child_name = absolute_name.rpartition(".")
            parent_module = TrussModuleLoader.import_truss_module(parent_name)
            path = parent_module.__spec__.submodule_search_locations

        spec = TrussModuleFinder.find_spec(absolute_name, path)
        module = importlib.util.module_from_spec(spec)
        sys.modules[absolute_name] = module
        spec.loader.exec_module(module)
        if path is not None:
            setattr(parent_module, child_name, module)
        return module


@contextmanager
def truss_module_loaded(
    truss_dir: str,
    truss_class_module_fullname: str,
    bundled_packages_dir_name: Optional[str] = None,
    external_packages_dirs: Optional[List[str]] = None,
):
    """
    Load a truss module: model or train.
    Args:
        truss_dir: Directory of truss
        truss_class_module_fullname: Full name of model or train module,
                                     e.g. 'model.model', 'train.train'
        bundled_packages_dir_name: name of the bundled packages directory
                                   if None then bundled packages are not loaded.
        external_packages_dirs: List of names of the external packages directories
                                   if None then external packages are not loaded.
    """
    try:
        truss_module_name = truss_class_module_fullname.split(".")[0]
        # Unload so that all modules can be freshly loaded.
        _unload_truss_modules(
            truss_dir,
            truss_module_name,
            bundled_packages_dir_name,
            external_packages_dirs,
        )
        TrussModuleFinder.set_model_truss_dirs(
            truss_dir,
            truss_module_name,
            bundled_packages_dir_name,
            external_packages_dirs,
        )
        class_module = TrussModuleLoader.import_truss_module(
            truss_class_module_fullname
        )
        yield class_module
    finally:
        TrussModuleFinder.remove_from_meta_path()


def _unload_truss_modules(
    truss_dir: str,
    truss_module_name: str,
    bundled_packages_dir_name: Optional[str] = None,
    external_packages_dirs: Optional[List[str]] = None,
) -> None:
    modules_to_unload = [truss_module_name]

    def _add_relative_dir_to_unload(dir_name: str):
        path = Path(truss_dir) / dir_name
        if path.exists():
            modules_to_unload.extend(_sub_dirnames(path))

    if bundled_packages_dir_name is not None:
        _add_relative_dir_to_unload(bundled_packages_dir_name)

    if external_packages_dirs is not None:
        for relative_dir in external_packages_dirs:
            _add_relative_dir_to_unload(relative_dir)
    _unload_top_level_modules(modules_to_unload)


def _unload_top_level_modules(module_names: List[str]) -> None:
    for module_name in module_names:
        _unload_top_level_module(module_name)


def _unload_top_level_module(module_name: str) -> None:
    if "." in module_name:
        raise ValueError(f"Expecting a top level module but found {module_name}")

    truss_module_submodules = [
        sys_module_name
        for sys_module_name in sys.modules
        if sys_module_name.startswith(module_name + ".")
    ]
    for truss_module_submodule in truss_module_submodules:
        del sys.modules[truss_module_submodule]
    if module_name in sys.modules:
        del sys.modules[module_name]


def _sub_dirnames(root_dir: Path) -> List[str]:
    return [path.name for path in root_dir.iterdir() if path.is_dir()]



import fnmatch
from pathlib import Path
from typing import List, Optional


def path_matches_any_pattern(path: Path, patterns: Optional[List[str]]) -> bool:
    if patterns is None:
        return False

    for pattern in patterns:
        if fnmatch.fnmatch(str(path), pattern):
            return True

    return False



import tempfile
from pathlib import Path

import pytest
import yaml
from truss.truss_config import (
    DEFAULT_CPU,
    DEFAULT_MEMORY,
    DEFAULT_USE_GPU,
    Accelerator,
    AcceleratorSpec,
    BaseImage,
    ModelCache,
    ModelRepo,
    Resources,
    Train,
    TrussConfig,
)
from truss.types import ModelFrameworkType


@pytest.mark.parametrize(
    "input_dict, expect_resources, output_dict",
    [
        (
            {},
            Resources(),
            {
                "cpu": DEFAULT_CPU,
                "memory": DEFAULT_MEMORY,
                "use_gpu": DEFAULT_USE_GPU,
                "accelerator": None,
            },
        ),
        (
            {"accelerator": None},
            Resources(),
            {
                "cpu": DEFAULT_CPU,
                "memory": DEFAULT_MEMORY,
                "use_gpu": DEFAULT_USE_GPU,
                "accelerator": None,
            },
        ),
        (
            {"accelerator": "V100"},
            Resources(accelerator=AcceleratorSpec(Accelerator.V100, 1), use_gpu=True),
            {
                "cpu": DEFAULT_CPU,
                "memory": DEFAULT_MEMORY,
                "use_gpu": True,
                "accelerator": "V100",
            },
        ),
        (
            {"accelerator": "T4:1"},
            Resources(accelerator=AcceleratorSpec(Accelerator.T4, 1), use_gpu=True),
            {
                "cpu": DEFAULT_CPU,
                "memory": DEFAULT_MEMORY,
                "use_gpu": True,
                "accelerator": "T4",
            },
        ),
        (
            {"accelerator": "A10G:4"},
            Resources(accelerator=AcceleratorSpec(Accelerator.A10G, 4), use_gpu=True),
            {
                "cpu": DEFAULT_CPU,
                "memory": DEFAULT_MEMORY,
                "use_gpu": True,
                "accelerator": "A10G:4",
            },
        ),
    ],
)
def test_parse_resources(input_dict, expect_resources, output_dict):
    parsed_result = Resources.from_dict(input_dict)
    assert parsed_result == expect_resources
    assert parsed_result.to_dict() == output_dict


@pytest.mark.parametrize(
    "input_str, expected_acc",
    [
        (None, AcceleratorSpec(None, 0)),
        ("T4", AcceleratorSpec(Accelerator.T4, 1)),
        ("A10G:4", AcceleratorSpec(Accelerator.A10G, 4)),
        ("A100:8", AcceleratorSpec(Accelerator.A100, 8)),
        ("H100", AcceleratorSpec(Accelerator.H100, 1)),
    ],
)
def test_acc_spec_from_str(input_str, expected_acc):
    assert AcceleratorSpec.from_str(input_str) == expected_acc


@pytest.mark.parametrize(
    "input_dict, expect_base_image, output_dict",
    [
        (
            {},
            BaseImage(),
            {
                "image": "",
                "python_executable_path": "",
            },
        ),
        (
            {"image": "custom_base_image", "python_executable_path": "/path/python"},
            BaseImage(image="custom_base_image", python_executable_path="/path/python"),
            {
                "image": "custom_base_image",
                "python_executable_path": "/path/python",
            },
        ),
    ],
)
def test_parse_base_image(input_dict, expect_base_image, output_dict):
    parsed_result = BaseImage.from_dict(input_dict)
    assert parsed_result == expect_base_image
    assert parsed_result.to_dict() == output_dict


def generate_default_config():
    config = {
        "environment_variables": {},
        "external_package_dirs": [],
        "model_metadata": {},
        "model_name": None,
        "python_version": "py39",
        "requirements": [],
        "resources": {
            "accelerator": None,
            "cpu": "1",
            "memory": "2Gi",
            "use_gpu": False,
        },
        "secrets": {},
        "system_packages": [],
    }
    return config


def test_default_config_not_crowded_end_to_end():
    config = TrussConfig(
        python_version="py39",
        requirements=[],
    )

    config_yaml = """environment_variables: {}
external_package_dirs: []
model_metadata: {}
model_name: null
python_version: py39
requirements: []
resources:
  accelerator: null
  cpu: '1'
  memory: 2Gi
  use_gpu: false
secrets: {}
system_packages: []
"""

    assert config_yaml.strip() == yaml.dump(config.to_dict(verbose=False)).strip()


@pytest.mark.parametrize(
    "model_framework",
    [ModelFrameworkType.CUSTOM, ModelFrameworkType.SKLEARN, ModelFrameworkType.PYTORCH],
)
def test_model_framework(model_framework):
    config = TrussConfig(
        python_version="py39",
        requirements=[],
        model_framework=model_framework,
    )

    new_config = generate_default_config()
    if model_framework == ModelFrameworkType.CUSTOM:
        assert new_config == config.to_dict(verbose=False)
    else:
        new_config["model_framework"] = model_framework.value
        assert new_config == config.to_dict(verbose=False)


def test_non_default_train():
    config = TrussConfig(
        python_version="py39",
        requirements=[],
        train=Train(resources=Resources(use_gpu=True, accelerator="A10G")),
    )

    updated_train = {
        "resources": {
            "accelerator": "A10G",
            "cpu": "1",
            "memory": "2Gi",
            "use_gpu": True,
        },
        "variables": {},
    }

    new_config = generate_default_config()
    new_config["train"] = updated_train

    assert new_config == config.to_dict(verbose=False)


def test_null_model_cache_key():
    config_yaml_dict = {"model_cache": None}
    with tempfile.NamedTemporaryFile(mode="w", delete=False) as tmp_file:
        yaml.safe_dump(config_yaml_dict, tmp_file)
    config = TrussConfig.from_yaml(Path(tmp_file.name))
    assert config.model_cache == ModelCache.from_list([])


def test_null_hf_cache_key():
    config_yaml_dict = {"hf_cache": None}
    with tempfile.NamedTemporaryFile(mode="w", delete=False) as tmp_file:
        yaml.safe_dump(config_yaml_dict, tmp_file)
    config = TrussConfig.from_yaml(Path(tmp_file.name))
    assert config.model_cache == ModelCache.from_list([])


def test_huggingface_cache_single_model_default_revision():
    config = TrussConfig(
        python_version="py39",
        requirements=[],
        model_cache=ModelCache(models=[ModelRepo("test/model")]),
    )

    new_config = generate_default_config()
    new_config["model_cache"] = [
        {
            "repo_id": "test/model",
        }
    ]

    assert new_config == config.to_dict(verbose=False)
    assert config.to_dict(verbose=True)["model_cache"][0].get("revision") is None


def test_huggingface_cache_single_model_non_default_revision():
    config = TrussConfig(
        python_version="py39",
        requirements=[],
        model_cache=ModelCache(models=[ModelRepo("test/model", "not-main")]),
    )

    assert config.to_dict(verbose=False)["model_cache"][0].get("revision") == "not-main"


def test_huggingface_cache_multiple_models_default_revision():
    config = TrussConfig(
        python_version="py39",
        requirements=[],
        model_cache=ModelCache(
            models=[
                ModelRepo("test/model1", "main"),
                ModelRepo("test/model2"),
            ]
        ),
    )

    new_config = generate_default_config()
    new_config["model_cache"] = [
        {"repo_id": "test/model1", "revision": "main"},
        {
            "repo_id": "test/model2",
        },
    ]

    assert new_config == config.to_dict(verbose=False)
    assert config.to_dict(verbose=True)["model_cache"][0].get("revision") == "main"
    assert config.to_dict(verbose=True)["model_cache"][1].get("revision") is None


def test_huggingface_cache_multiple_models_mixed_revision():
    config = TrussConfig(
        python_version="py39",
        requirements=[],
        model_cache=ModelCache(
            models=[
                ModelRepo("test/model1"),
                ModelRepo("test/model2", "not-main2"),
            ]
        ),
    )

    new_config = generate_default_config()
    new_config["model_cache"] = [
        {
            "repo_id": "test/model1",
        },
        {"repo_id": "test/model2", "revision": "not-main2"},
    ]

    assert new_config == config.to_dict(verbose=False)
    assert config.to_dict(verbose=True)["model_cache"][0].get("revision") is None
    assert config.to_dict(verbose=True)["model_cache"][1].get("revision") == "not-main2"


def test_empty_config():
    config = TrussConfig()
    new_config = generate_default_config()

    assert new_config == config.to_dict(verbose=False)


def test_from_yaml():
    yaml_path = Path("test.yaml")
    data = {"description": "this is a test"}
    with yaml_path.open("w") as yaml_file:
        yaml.safe_dump(data, yaml_file)

    result = TrussConfig.from_yaml(yaml_path)

    assert result.description == "this is a test"

    yaml_path.unlink()


def test_from_yaml_empty():
    yaml_path = Path("test.yaml")
    data = {}
    with yaml_path.open("w") as yaml_file:
        yaml.safe_dump(data, yaml_file)

    result = TrussConfig.from_yaml(yaml_path)

    # test some attributes (should be default)
    assert result.description is None
    assert result.spec_version == "2.0"
    assert result.bundled_packages_dir == "packages"

    yaml_path.unlink()
